#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.


from mitosheet.step_performers.import_steps.simple_import import SimpleImportStepPerformer


CLEAR_EVENT = 'clear'
CLEAR_PARAMS = []

def execute_clear_update(steps_manager):
    steps_manager.execute_clear()

CLEAR_UPDATE = {
    'event_type': CLEAR_EVENT,
    'params': CLEAR_PARAMS,
    'execute': execute_clear_update
}