#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains tests to make sure that the mito analytics test is
performing correctly
"""

# TODO: add tests for anonymization